#ifndef __ECC_H__
#define __ECC_H__

#include "msp430.h"
#include "msp430def.h"

#define true 		1
#define false 		0


//flag if error counting after decoding is needed
#define COUNT_ERR	    			true
//defines the max size of a decoded payload
#define ECC_MAX_DEC_PAYLOAD_SIZE	104
//defines the max size of a encoded payload can have (limit to 52 bytes to encode (incl CRC bytes))
#define ECC_MAX_PAYLOAD_SIZE		156

typedef int bool;

void ECC_run_test_program();

void ECC_run_test_program_II();

void ECC_setMode(uint8_t mode);

bool ECC_encode(char* in, uint16_t in_s, char* out, uint16_t out_s);

bool ECC_enc_ham();

void ECC_mult_with_mat(uint8_t* input_vec, uint8_t input_vec_s, uint8_t* matrix, uint8_t* output_vec, uint8_t output_vec_s);

void ECC_doHammingEncoding(char* in, uint16_t in_s, char* out, uint16_t out_s);

bool ECC_enc_DECTED();

uint8_t ECC_calcBufferSize(uint8_t size, bool isEncode);

uint8_t ECC_decode(char* in, uint16_t in_s, char* out, uint16_t out_s);

void ECC_printDecoderState();

void ECC_initErrStats();

uint8_t ECC_dec_ham();

void ECC_doHammingDecoding(char* in, uint16_t in_s, char* out, uint16_t out_s);

uint8_t ECC_dec_DECTED();

uint8_t ECC_dec_bch();

bool ECC_convertBit2Byte(uint8_t* bitseq, uint8_t bitseq_s, uint8_t* res_byte);

void ECC_printBits(char* byte);

void ECC_setErrStats(uint8_t err);

void ECC_doHammingCorrection(uint8_t* data, uint8_t* syn);

void ECC_doDECTEDCorrection(uint8_t* data, uint8_t data_s, uint8_t* syn);

bool ECC_swapBit(uint8_t* data, uint8_t data_s, uint8_t pos);

void ECC_and64(char* l1, char* l2, char* res);

int ECC_multGF64(uint8_t a, uint8_t b);

int ECC_findAlpha(uint8_t bits);

uint8_t ECC_dec_rep(uint8_t r, uint8_t k);

bool ECC_calcSyndromes(uint8_t* corr_data_block, uint8_t* synd);

void ECC_doBerleMass(uint8_t* synd, int* err_poly);

uint8_t ECC_doChienSearch(int* err_poly, uint8_t* error_bits);

void ECC_xor64(char* l1, char* l2, char* res);

void ECC_rshift64(char* l1, uint8_t shift);

void ECC_lshift64(char* l1, uint8_t shift);

int ECC_readSizeBits(uint8_t* data);

bool ECC_findRoot(uint8_t* roots, uint8_t root);

void ECC_doDivisionEncoding(char* dataBytes, char* enc_dataBytes, uint8_t usedPayloadBytes);

void ECC_convertBytes2Bits(char* inBuf, uint8_t inBuf_s, char* outBuf, uint16_t out_s);

bool ECC_enc_bch();

bool ECC_enc_rep(uint8_t r, uint8_t k);

void ECC_setSizeBits(uint8_t* bit_ptr, uint8_t num_bytes_used);

void ECC_doPolyDiv(uint8_t* data, uint8_t* rem);

void ECC_convertBits2Bytes(uint8_t* in, uint16_t in_s, uint8_t* out, uint8_t out_s);

void ECC_printBytesRow(char* buf, uint8_t buf_s);

void ECC_printErrStats();

/**
 * Stores the Generator Matrix G for Hamming (7,4) (SECDED) (found in wikipedia (systematic))
 */
static uint8_t const generator_ham74[28] = {
		1, 0, 0, 0, 1, 1, 0,
		0, 1, 0, 0, 1, 0, 1,
		0, 0, 1 ,0, 0, 1, 1,
		0, 0, 0, 1, 1, 1, 1
};

/**
 * Stores the Parity check Matrix H for Hamming (7,4) (SECDED) (found in wikipedia (systematic))
 */
static uint8_t const parityCheck_ham74[21] = {
		1, 1, 0,
		1, 0, 1,
		0, 1, 1,
		1, 1, 1,
		1, 0, 0,
		0, 1, 0,
		0, 0, 1
};

/**
 * Stores the Generator Matrix G for Hamming (1
 * 6,8) (DECTED) (found in T. Aaron Gulliver and Vijay K. Bhargava)
 */
static uint8_t const generator_ham168[128] = {
		1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   1,   0,   1,   0,
		0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   1,   0,   1,
		0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,   1,   0,   1,   0,
		0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,   1,   0,   1,
		0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   1,   1,   0,
		0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   1,   1,
		0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   1,   0,   1,   0,   0,   1,
		0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   0,   1,   0,   1,   0,   0
};

/**
 * Stores the Parity check Matrix H for Hamming (16,8) (DECTED) (found in T. Aaron Gulliver and Vijay K. Bhargava)
 */
static uint8_t const parityCheck_ham168[128] = {
		0,   1,   1,   0,   1,   0,   1,   0,
		0,   0,   1,   1,   0,   1,   0,   1,
		1,   0,   0,   1,   1,   0,   1,   0,
		0,   1,   0,   0,   1,   1,   0,   1,
		1,   0,   1,   0,   0,   1,   1,   0,
		0,   1,   0,   1,   0,   0,   1,   1,
		1,   0,   1,   0,   1,   0,   0,   1,
		1,   1,   0,   1,   0,   1,   0,   0,
		1,   0,   0,   0,   0,   0,   0,   0,
		0,   1,   0,   0,   0,   0,   0,   0,
		0,   0,   1,   0,   0,   0,   0,   0,
		0,   0,   0,   1,   0,   0,   0,   0,
		0,   0,   0,   0,   1,   0,   0,   0,
		0,   0,   0,   0,   0,   1,   0,   0,
		0,   0,   0,   0,   0,   0,   1,   0,
		0,   0,   0,   0,   0,   0,   0,   1
};

static uint8_t const singleBitError_patterns_DECTED[16] = {
		// 106, 53, 154, 77, 166, 83, 169, 212, 1,	2, 4, 8, 16, 32, 64, 128 // was wrong due to new version of conversion method
		86, 172, 89, 178, 101, 202, 149, 43, 128, 64, 32, 16, 8, 4, 2, 1
};

/**
 * Stores the identifier sums for all 2 bit errors in the data block (C^T) (DECTED (16,8))
 * below diag: parity bit errors
 * above diag: data bit errors
 */
static uint8_t const data2BitErrorSameBlock_patterns_DECTED[8][8] = {
	{0,		250,	15,		228,	51,		156,	195,	125},
	{3,		0,		245,	30,		201,	102,	57,		135},
	{5,		6,		0,		235,	60,		147,	204,	114},
	{9,		10,		12,		0,		215,	120,	39,		153},
	{17,	18,		20,		24,		0,		175,	240,	78},
	{33,	34,		36,		40,		48,		0,		95,		225},
	{65,	66,		68,		72,		80,		96,		0,		190},
	{129,	6,		132,	136,	144,	160,	192,	0}
};

/**
 * Stores all the 2 Bit Error patterns where one error is in C^T and one in I.
 * (x,y) stores the sum of the combination of the column x of C^t and column y of I
 */
static uint8_t const parityAndData2BitError_patterns_DECTED[8][8] = {

		{87, 	84, 	82, 	94, 	70, 	118, 	22, 	214},
		{173, 	174, 	168, 	164, 	188, 	140, 	236, 	44},
		{88, 	91, 	93, 	81, 	73, 	121, 	25, 	217},
		{179, 	176, 	182, 	186, 	162, 	146, 	242, 	50},
		{100, 	103, 	97, 	109, 	117, 	69, 	37, 	229},
		{203, 	200, 	206, 	194, 	218, 	234, 	138, 	74},
		{148, 	151, 	145, 	157, 	133, 	181, 	213, 	21},
		{42, 	41, 	47, 	35, 	59, 	11, 	107, 	171}
};
// used to fasten bit 2 byte conversion
static const uint8_t potencies[8] = {1, 2, 4, 8, 16, 32, 64, 128};

/**
 * Contains the generator polynom for BCH (63,57) encoded as a long long (64-Bit)
 * bit sequence: 1000011
 */
//static unsigned long long const gen_poly_BCH_63_57 = 67;
static uint8_t const gen_poly_BCH_63_57[7] = {1,1,0,0,0,0,1};

/**
 * Contains the generator polynom for BCH (63,51) encoded as a long long (64-Bit)
 * bit sequence: 1010100111001
 */
//static unsigned long long const gen_poly_BCH_63_51 = 5433;
static uint8_t const gen_poly_BCH_63_51[13] = {1,0,0,1,1,1,0,0,1,0,1,0,1};

/**
 * Contains the generator polynom for BCH (63,45) encoded as a long long (64-Bit)
 * bit sequence: 1111000001011001111
 */
//static unsigned long long const gen_poly_BCH_63_45 = 492239;
static uint8_t const gen_poly_BCH_63_45[19] = {1,1,1,1,0,0,1,1,0,1,0,0,0,0,0,1,1,1,1};

/**
 * Contains the generator polynom for BCH (63,39) encoded as a long long (64-Bit)
 * bit sequence: 1110110110010011101110111
 */
//static unsigned long long const gen_poly_BCH_63_39 = 31139703;
static uint8_t const gen_poly_BCH_63_39[25] = {1,1,1,0,1,1,1,0,1,1,1,0,0,1,0,0,1,1,0,1,1,0,1,1,1};

/**
 * Contains the generator polynom for BCH (63,36) encoded as a long long (64-Bit)
 * bit sequence: 1000011011101000000100010011
 */
//static unsigned long long const gen_poly_BCH_63_36 = 141459731;
static uint8_t const gen_poly_BCH_63_36[28] = {1,1,0,0,1,0,0,0,1,0,0,0,0,0,0,1,0,1,1,1,0,1,1,0,0,0,0,1};


/**
 * Contains the binary representation of the powers of alpha (=2 in GF(2^m)) in GF(2^6) = GF(64)
 * The representation is valid for all gen polys used in this code
 * This alpha representations are used to calculate the syndrome values of the BCH codes
 */
static const uint8_t alpha_to[63] = {
		// generated by http://www.eccpage.com/bch3.c
		1, 2, 4, 8, 16, 32, 3, 6, 12, 24, 48, 35, 5, 10, 20, 40, 19, 38, 15, 30, 60, 59, 53, 41, 17, 34, 7, 14, 28, 56, 51, 37, 9, 18, 36, 11, 22, 44, 27, 54, 47, 29, 58, 55, 45, 25, 50, 39, 13, 26, 52, 43, 21, 42, 23, 46, 31, 62, 63, 61, 57, 49, 33
};

/**
 * table which knows the num of 1's set in every byte
 */
static const uint8_t BitsSetTable256[] =
{
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};




#endif
